package game;


public class AttemptShot {

	public static boolean attemptShot(int sx, int sy, int tx, int ty, GameMap map){

		/** Various variables for collision calculations
		 * txdb etc exist so that we can add 0.5 to get the centre of a cell rather than the top left
		 */

		double txdb = tx + 0.5;
		double tydb = ty + 0.5;
		double sxdb = sx + 0.5;
		double sydb = sy + 0.5;	
		double changeX = txdb-sxdb;
		double changeY = tydb-sydb;		
		double shotGradient = 0;
		int counter = 0;

		/**
		 * This allows us to avoid divide by 0 errors
		 */

		if (changeX !=0 && changeY != 0) {shotGradient = changeY/changeX;}		
		double yIntercept = - ((shotGradient * sxdb) - sydb);

		/**
		 * Calculations for euclidean distance
		 */
		double shotDistance = ((txdb-sxdb)*(txdb-sxdb))+((tydb-sydb)*(tydb-sydb));
		double rootShotDistance = (double) Math.sqrt(shotDistance);
		int intRootShotDistance = (int) Math.ceil(rootShotDistance);

		/**
		 * This checks that a cell is within a circular radius of the selected unit
		 * A cell must be completely within the circular radius for it to be considered legal
		 */

		if (intRootShotDistance > map.getUnitInNode((int)sxdb,(int)sydb).getShootDistance()){
			return false;
		}

		if (map.getUnitInNode(sx, sy).getimageloc() == 10 || map.getUnitInNode(sx, sy).getimageloc() == 11){
			//System.out.println("Grenadier found...");
			if (intRootShotDistance <= 1){
				return false;
			}
			else return true;
		}

		/**
		 * These loops consider shots that are completely horizontal or vertical
		 */

		if (changeX == 0 || changeY == 0){

			if (changeX == 0 && changeY < 0) {
				for (int i = sy; i >= ty; i--){
					if (map.tiletypes[sx][i] == 3){						
						return false;
					}
					if (map.tiletypes[sx][i] == 1){						
						counter = counter + 1;
					}
				}
				
				int temprange = map.getUnitInNode((int)sxdb,(int)sydb).getShootDistance();
				int tempdistance = intRootShotDistance;
				int totaldist = tempdistance + counter;								
				if (totaldist > temprange){
					return false;

				}
			}

			else if (changeX == 0 && changeY > 0) {
				for (int i = sy; i <= ty; i++){
					if (map.tiletypes[sx][i] == 3){						
						return false;
					}
					if (map.tiletypes[sx][i] == 1){						
						counter = counter + 1;
					}
					
				}
				int temprange = map.getUnitInNode((int)sxdb,(int)sydb).getShootDistance();
				int tempdistance = intRootShotDistance;
				int totaldist = tempdistance + counter;								
				if (totaldist > temprange){
					return false;

				}
			}

			else if (changeY == 0 && changeX > 0) {
				for (int i = sx; i <= tx; i++){
					if (map.tiletypes[i][sy] == 3){
						return false;
					}
					if (map.tiletypes[i][sy] == 1){						
						counter = counter + 1;
					}
				}
				int temprange = map.getUnitInNode((int)sxdb,(int)sydb).getShootDistance();
				int tempdistance = intRootShotDistance;
				int totaldist = tempdistance + counter;								
				if (totaldist > temprange){
					return false;

				}
			}

			else if (changeY == 0 && changeX < 0) {
				for (int i = sx; i >= tx; i--){
					if (map.tiletypes[i][sy] == 3){				
						return false;
					}
					if (map.tiletypes[i][sy] == 1){						
						counter = counter + 1;
					}
				}
				int temprange = map.getUnitInNode((int)sxdb,(int)sydb).getShootDistance();
				int tempdistance = intRootShotDistance;
				int totaldist = tempdistance + counter;								
				if (totaldist > temprange){
					return false;

				}
			}	
		}

		/** 
		 * These for loops consider diagonal attacks: one for each possible diagonal direction
		 */

		else if (sxdb < txdb && sydb < tydb) {		
			for (int i =  sx; i <= tx; i++){
				for (int j = sy; j <= ty; j++){

					//if a tree
					if (map.tiletypes[i][j] == 1){
						double testY = ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {

							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){

							//add one to the tally of trees found
							counter = counter + 1;
						}
					}

					if (map.tiletypes[i][j] == 3){

						double testY = ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {

							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						//if both conditions are false there's a wall in the way of our shot
						if (cond1 == false && cond2 == false){												
							return false;
						}
					}
				}
			}	
			
			//at this point we've looped every tile
			//if there's no wall in the way of the shot, add the number of trees in line of the shot to
			//the distance of the shot, and check its still in range
			int temprange = map.getUnitInNode((int)sxdb,(int)sydb).getShootDistance();
			int tempdistance = intRootShotDistance;
			int totaldist = tempdistance + counter;								
			if (totaldist > temprange){
				return false;

			}
		}

		else if (sxdb < txdb && sydb > tydb) {			
			for (int i =  sx; i <= tx; i++){
				for (int j =  sy; j >= ty; j--){

					if (map.tiletypes[i][j] == 1){
						double testY = ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {

							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){
							//add one to the tally of trees found

							counter = counter + 1;
						}
					}

					if (map.tiletypes[i][j] == 3){

						double testY =  ((shotGradient * i) + yIntercept);
						double testYPlus =  ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {

							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){
							return false;
						}

						
					}
				}
			}
			int temprange = map.getUnitInNode((int)sxdb,(int)sydb).getShootDistance();
			int tempdistance = intRootShotDistance;
			int totaldist = tempdistance + counter;								
			if (totaldist > temprange){
				return false;

			}
		}

		else if (sxdb > txdb && sydb > tydb) {		
			for (int i =  sx; i >= tx; i--){
				for (int j =  sy; j >= ty; j--){

					if (map.tiletypes[i][j] == 1){
						double testY = ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {

							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){

							//add one to the tally of trees found
							counter = counter + 1;
						}
					}

					if (map.tiletypes[i][j] == 3){

						double testY =  ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);					
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {

							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){
							return false;
						}
					}
				}
			}	
			int temprange = map.getUnitInNode((int)sxdb,(int)sydb).getShootDistance();
			int tempdistance = intRootShotDistance;
			int totaldist = tempdistance + counter;								
			if (totaldist > temprange){
				return false;

			}
		}

		else if (sxdb > txdb && sydb < tydb) {	
			for (int i =  sx; i >= tx; i--){
				for (int j =  sy; j <= ty; j++){

					if (map.tiletypes[i][j] == 1){
						double testY = ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {

							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){

							//add one to the tally of trees found
							counter = counter + 1;
						}
					}

					if (map.tiletypes[i][j] == 3){					
						double testY = ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {
							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; } //testyplus
							if (testYPlus <= j) { cond2 = true; } //testy
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){
							return false;
						}
					}
				}
			}
		
			int temprange = map.getUnitInNode((int)sxdb,(int)sydb).getShootDistance();
			int tempdistance = intRootShotDistance;
			int totaldist = tempdistance + counter;								
			if (totaldist > temprange){
				return false;

			}
		}
		return true;
	}

	/**
	 * This is an alternative version of the attempt shot range used during the generation of potential moves
	 * When a new turn is created where a player moves, then fires, no unit is actually moved until a new state is created
	 * As such there is no unit in sx/sy, so we need an extra argument, shotRange, in order to complete the calculations
	 */

	public static boolean testShot(int sx, int sy, int tx, int ty, GameMap map, int shotRange, int oldsx, int oldsy){

		double txdb = tx + 0.5;
		double tydb = ty + 0.5;
		double sxdb = sx + 0.5;
		double sydb = sy + 0.5;	
		double changeX = txdb-sxdb;
		double changeY = tydb-sydb;	
		double shotGradient = 0;
		int counter = 0;
		if (changeX !=0 && changeY != 0) {shotGradient = changeY/changeX;}
		double yIntercept = - ((shotGradient * sxdb) - sydb);

		double shotDistance = ((txdb-sxdb)*(txdb-sxdb))+((tydb-sydb)*(tydb-sydb));
		double rootShotDistance = (double) Math.sqrt(shotDistance);
		int intRootShotDistance = (int) Math.ceil(rootShotDistance);

		if (intRootShotDistance > shotRange){		
			return false;
		}

		if (map.getUnitInNode(oldsx, oldsy).getimageloc() == 10 || map.getUnitInNode(oldsx, oldsy).getimageloc() == 11){
			//System.out.println("Grenadier found...");
			if (intRootShotDistance < 1){
				return false;
			}
			else return true;
		}

		if (changeX == 0 || changeY == 0){
			if (changeX == 0 && changeY < 0) {
				for (int i = sy; i >= ty; i--){
					if (map.tiletypes[sx][i] == 3){
						return false;
					}
					if (map.tiletypes[sx][i] == 1){						
						counter = counter + 1;
					}
				}
				
				int temprange = map.getUnitInNode((int)oldsx,(int)oldsy).getShootDistance();
				int tempdistance = intRootShotDistance;
				int totaldist = tempdistance + counter;								
				if (totaldist > temprange){
					return false;

				}
			}

			else if (changeX == 0 && changeY > 0) {
				for (int i = sy; i <= ty; i++){
					if (map.tiletypes[sx][i] == 3){
						return false;
					}
					if (map.tiletypes[sx][i] == 1){						
						counter = counter + 1;
					}
				}
				int temprange = map.getUnitInNode((int)oldsx,(int)oldsy).getShootDistance();
				int tempdistance = intRootShotDistance;
				int totaldist = tempdistance + counter;								
				if (totaldist > temprange){
					return false;

				}
			}

			else if (changeY == 0 && changeX > 0) {
				for (int i = sx; i <= tx; i++){
					if (map.tiletypes[i][sy] == 3){
						return false;
					}
					if (map.tiletypes[i][sy] == 1){						
						counter = counter + 1;
					}
				}
				int temprange = map.getUnitInNode((int)oldsx,(int)oldsy).getShootDistance();
				int tempdistance = intRootShotDistance;
				int totaldist = tempdistance + counter;								
				if (totaldist > temprange){
					return false;

				}
			}

			else if (changeY == 0 && changeX < 0) {
				for (int i = sx; i >= tx; i--){
					if (map.tiletypes[i][sy] == 3){
						return false;
					}
					if (map.tiletypes[i][sy] == 1){						
						counter = counter + 1;
					}
				}
				int temprange = map.getUnitInNode((int)oldsx,(int)oldsy).getShootDistance();
				int tempdistance = intRootShotDistance;
				int totaldist = tempdistance + counter;								
				if (totaldist > temprange){
					return false;

				}
			}	
		}

		else if (sxdb < txdb && sydb < tydb) {	
			for (int i =  sx; i <= tx; i++){
				for (int j = sy; j <= ty; j++){

					if (map.tiletypes[i][j] == 1){
						double testY = ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {

							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){

							//add one to the tally of trees found
							counter = counter + 1;
						}
					}

					if (map.tiletypes[i][j] == 3){

						double testY = ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {
							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){
							return false;
						}
					}
				}
			}	
			
			int temprange = map.getUnitInNode((int)oldsx,(int)oldsy).getShootDistance();
			int tempdistance = intRootShotDistance;
			int totaldist = tempdistance + counter;								
			if (totaldist > temprange){
				return false;
			}
		}

		else if (sxdb < txdb && sydb > tydb) {
			for (int i =  sx; i <= tx; i++){
				for (int j =  sy; j >= ty; j--){

					if (map.tiletypes[i][j] == 1){
						double testY = ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {

							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){

							//add one to the tally of trees found
							counter = counter + 1;
						}
					}

					if (map.tiletypes[i][j] == 3){

						double testY =  ((shotGradient * i) + yIntercept);
						double testYPlus =  ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {

							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){
							return false;
						}
					}
				}
			}	
			
			int temprange = map.getUnitInNode((int)oldsx,(int)oldsy).getShootDistance();
			int tempdistance = intRootShotDistance;
			int totaldist = tempdistance + counter;								
			if (totaldist > temprange){
				return false;
			}
		}
		else if (sxdb > txdb && sydb > tydb) {
			for (int i =  sx; i >= tx; i--){
				for (int j =  sy; j >= ty; j--){

					if (map.tiletypes[i][j] == 1){
						double testY = ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {

							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){

							//add one to the tally of trees found
							counter = counter + 1;
						}
					}

					if (map.tiletypes[i][j] == 3){

						double testY =  ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {
							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){
							return false;
						}
					}
				}
			}
			
			int temprange = map.getUnitInNode((int)oldsx,(int)oldsy).getShootDistance();
			int tempdistance = intRootShotDistance;
			int totaldist = tempdistance + counter;								
			if (totaldist > temprange){
				return false;
			}
		}

		else if (sxdb > txdb && sydb < tydb) {
			for (int i =  sx; i >= tx; i--){
				for (int j =  sy; j <= ty; j++){

					if (map.tiletypes[i][j] == 1){
						double testY = ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {

							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;

							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){

							//add one to the tally of trees found
							counter = counter + 1;
						}
					}

					if (map.tiletypes[i][j] == 3){

						double testY = ((shotGradient * i) + yIntercept);
						double testYPlus = ((shotGradient * (i+1)) + yIntercept);
						boolean cond1 = false;
						boolean cond2 = false;

						if (testY < testYPlus) {
							if (testY >= (j+1)) { cond1 = true; }
							if (testYPlus <= j) { cond2 = true; }
						}

						else if (testY > testYPlus) {
							double temp = testY;
							testY = testYPlus;	
							testYPlus = temp;
							if (testY >= (j+1)) { cond1 = true; } //testyplus
							if (testYPlus <= j) { cond2 = true; } //testy
						}

						else if (testY == testYPlus) {
							if (testY > (j+1)) { cond1 = true; }
							else if (testYPlus < (j)) { cond2 = true; }
						}

						if (cond1 == false && cond2 == false){
							return false;
						}
					}
				}
			}		
			
			int temprange = map.getUnitInNode((int)oldsx,(int)oldsy).getShootDistance();
			int tempdistance = intRootShotDistance;
			int totaldist = tempdistance + counter;								
			if (totaldist > temprange){
				return false;
			}
		}
		return true;
	}
}

