package game;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.EventQueue;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Window.Type;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import javax.imageio.ImageIO;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.border.EmptyBorder;

import pathfinder.AStar;
import pathfinder.Path;
import pathfinder.PathFinderInterface;

import players.AIFactory;
import players.PlayerInterface;
import units.UnitInterface;

import javax.swing.JLabel;
import javax.swing.JLayeredPane;
import javax.swing.JButton;
import javax.swing.JToggleButton;

import game.GameMap;
import game.GameInterface.MapSelect;

public class Game extends JFrame {

	/**
	 * Various variables used by the game 
	 * @tiles is an array that holds the tile images
	 * @buffer holds the game as it is computed so it can be drawn as a single object
	 * @turnID 0 is red 1 is blue
	 */

	private static final long serialVersionUID = -359152097369721457L;
	
	private GameMap map; 

	public static int initialRed;
	public static int initialBlue;
	
	public static PathFinderInterface finder;
	public Path path; 
	private Image[] tiles = new Image[13]; 
	private Image buffer; 
	public int turnCount = 0;
	public static int maxTurns = -1;

	private int turnID = 0; 
	private boolean showShotDistance;
	private JPanel contentPane;

	private players.HumanPlayer player1 = new players.HumanPlayer(0);
	private players.PlayerInterface player2;
	private AIFactory factory = new AIFactory();
	public static players.PlayerInterface currentPlayer;
	
	/**
	 * Stats the game tracks 
	 */

	public static int globalturncount = 0;
	public static int globalredkills = 0;
	public static int globalbluekills = 0;
	public static int globalredwins = 0;
	public static int globalbluewins = 0;

	public Game(boolean useEditorMap) {
		
		player2 = factory.createAI(GameInterface.AIChoice);
		
		if (useEditorMap == false){
			this.map = new GameMap();
		}
		
		else if (useEditorMap == true){

			this.map = new GameMap(MapEditor.savedMap);
		}
		
		initialRed = map.redUnitCount;
		initialBlue = map.blueUnitCount;

		setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		int windowX = ((map.getWidth() * 32) + 100);
		int windowY = ((map.getHeight() * 32) + 150);
		setBounds(100, 100, 450, 300);
		contentPane = new JPanel();
		contentPane.setBorder(new EmptyBorder(5, 5, 5, 5));
		contentPane.setLayout(new BorderLayout(0, 0));
		setContentPane(contentPane);
		setSize(windowX,windowY);
		setResizable(true);
		setVisible(true);
		JLayeredPane layeredPane = new JLayeredPane();
		contentPane.add(layeredPane, BorderLayout.CENTER);
		gameGraphics gGraphics = new gameGraphics();
		gGraphics.setSize(1024,1024);
		gGraphics.setOpaque(false);
		layeredPane.add(gGraphics, new Integer(0)); 

		/** 
		 * This is a button that ends the current turn and passes control to the other play
		 */

		JButton btnNewButton = new JButton("End Turn");
		layeredPane.setLayer(btnNewButton, 1);
		btnNewButton.setBounds(10, 11, 89, 23);
		btnNewButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				
				currentPlayer.swapPlayer(instanceRetriever());
				repaint(0);
			}
		});
		layeredPane.add(btnNewButton);

		/**
		 * This is a button that highlights the current player's selected unit's attack range
		 */

		JToggleButton tglbtnShowAttackRange = new JToggleButton("Show Attack Range");
		layeredPane.setLayer(tglbtnShowAttackRange, 1);
		tglbtnShowAttackRange.setBounds(109, 11, 136, 23);
		tglbtnShowAttackRange.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if (showShotDistance == false) {showShotDistance = true;}
				else if(showShotDistance == true) {showShotDistance = false;}
				for (int i = 0; i < map.getWidth(); i++){
					for (int j = 0; j < map.getHeight(); j++){
						if (getCurrentPlayer().getSelectedx() != -1){
							map.legalShots[i][j] = true;
							if ((AttemptShot.attemptShot(getCurrentPlayer().getSelectedx(), getCurrentPlayer().getSelectedy(), i, j, map) == true)){
								map.legalShots[i][j] = false;
							}

							if (map.tiletypes[i][j] == 3 || map.tiletypes[i][j] == 2){
								map.legalShots[i][j] = true;
							}
						}
					}
				}
				repaint(0);	
				System.out.println(map.legalShots);		
			}
		});
		layeredPane.add(tglbtnShowAttackRange);

		/**
		 * This loads images from the res folder into memory 
		 */

		try {
			tiles[GameMap.GRASS] = ImageIO.read(getResource("res/tile/GRASS.png"));
			tiles[GameMap.TREE] = ImageIO.read(getResource("res/tile/TREE.png"));
			tiles[GameMap.WATER] = ImageIO.read(getResource("res/tile/WATER.png"));
			tiles[GameMap.WALL] = ImageIO.read(getResource("res/tile/WALL.png"));
			tiles[GameMap.SOLDIER_RED] = ImageIO.read(getResource("res/unit/SOLDIER_RED.png"));
			tiles[GameMap.SOLDIER_BLUE] = ImageIO.read(getResource("res/unit/SOLDIER_BLUE.png"));
			tiles[GameMap.COMM_RED] = ImageIO.read(getResource("res/unit/COMM_RED.png"));
			tiles[GameMap.COMM_BLUE] = ImageIO.read(getResource("res/unit/COMM_BLUE.png"));
			tiles[GameMap.SNIPER_RED] = ImageIO.read(getResource("res/unit/SNIPER_RED.png"));
			tiles[GameMap.SNIPER_BLUE] = ImageIO.read(getResource("res/unit/SNIPER_BLUE.png"));	
			tiles[GameMap.GREN_RED] = ImageIO.read(getResource("res/unit/GRENADIER_RED.png"));
			tiles[GameMap.GREN_BLUE] = ImageIO.read(getResource("res/unit/GRENADIER_BLUE.png"));	
		}

		catch (IOException e) {
			System.err.println("Failed to load resources: "+e.getMessage());
			System.exit(0);
		}

		/** 
		 * Initial allocations for current player ID
		 */

		if (turnID == 0) { setCurrentPlayer(getPlayer1()); }
		if (turnID == 1) { setCurrentPlayer(getPlayer2()); }

		/**
		 * Adds a mouse listener for click events
		 * Code that handles click events is at end of class in handleMousePressed
		 */

		addMouseListener(new MouseAdapter() {
			public void mousePressed(MouseEvent e) {
				if (getCurrentPlayer().isAI() == false) {
					handleMousePressed(e.getX(), e.getY(), getCurrentPlayer(), instanceRetriever().getMap());
					//the pathfinder needs to be created here because the calculation on move distance is in the mouse
					//listener
					finder = new AStar(getMap(), getCurrentPlayer().getMoverange(), false);
				}
			}
		});

		/**
		 * Adds a mouse listener that handles mouse movement
		 * Code is in handleMouseMoved at end of class
		 */

		addMouseMotionListener(new MouseMotionListener() {
			public void mouseDragged(MouseEvent e) {
			}
			public void mouseMoved(MouseEvent e) {
				if (getCurrentPlayer().isAI() == false) {
					handleMouseMoved(e.getX(), e.getY(), getCurrentPlayer());
				}
			}
		});

		/**
		 * Kills the process if the window is closed
		 */
		addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {

			}
		});	

		/**
		 * Essentially begins the game by calling chooseMove in the current player's class
		 * Works for both an AI player and a human player
		 */
	
		getCurrentPlayer().chooseMove(this);

	}

	/**
	 * This inputstream is used when grabbing image files
	 */

	private InputStream getResource(String ref) throws IOException {
		return new FileInputStream(ref);
	}

	/**
	 * This class draws the game graphics to a buffer so they can be drawn as one image with repaint()
	 */

	class gameGraphics extends JPanel {

		private static final long serialVersionUID = -291953696912289802L;
		public void paint(Graphics graphics) {	

			if (buffer == null) {
				buffer = new BufferedImage(1024, 1024, BufferedImage.TYPE_INT_ARGB);			
			}
			Graphics g = buffer.getGraphics();

			g.clearRect(0,0,1024,1024);
			g.translate(50, 50);

			/**
			 * Loops through the map and draws the relevant tile
			 */

			for (int x=0;x<map.getWidth();x++) {
				for (int y=0;y<map.getHeight();y++) {
					g.drawImage(tiles[map.getTile(x, y)],x*32,y*32,null);
					if (map.getUnitInNode(x, y) != null && map.getUnitInNode(x, y).isAlive() == true) {
						g.drawImage(tiles[map.getUnitInNode(x, y).getimageloc()],x*32,y*32,null);
					} else {
						if (path != null) {
							if (path.contains(x, y)) {
								g.setColor(Color.blue);
								g.fillRect((x*32)+8, (y*32)+8,14,14);
							}
						}	
					}				
				}				
			}

			if (showShotDistance == true){
				for (int i = 0; i < map.getWidth(); i++){
					for (int j = 0; j < map.getHeight(); j++){
						if (map.legalShots[i][j] != true){
							g.setColor(Color.red);
							g.fillRect((i*32)+11, (j*32)+11,8,8);
						}
					}
				}
			}

			/**
			 * Draws a box around the selected unit
			 */

			if (getCurrentPlayer().getSelectedx() != -1) {
				g.setColor(Color.black);
				g.drawRect(getCurrentPlayer().getSelectedx()*32, getCurrentPlayer().getSelectedy()*32, 30, 30);				
				g.drawRect((getCurrentPlayer().getSelectedx()*32)-4, (getCurrentPlayer().getSelectedy()*32)-4, 38, 38);
				g.setColor(Color.white);
				g.drawRect((getCurrentPlayer().getSelectedx()*32)-2, (getCurrentPlayer().getSelectedy()*32)-2, 34, 34);
			}
			graphics.drawImage(buffer, 0, 0, null);
		}	
	}

	/**
	 * This class is used by the AI to generate all the possible moves it can make
	 * @param input: this is the input gamestate
	 */

	public static List<Turn> getLegalMoves(GameMap input, boolean largeCull){
		
	

		GameMap inputCopy = new GameMap(input);
		Turn newTurn = new Turn(-1,-1,-1,-1,-1,-1,-1,-1,false,false,false);
		List<Turn> candidateMoves;
		List<Turn> firedMoves;
		List<Turn> nonFiredMoves;
		List<Turn> furtherMoves;
		candidateMoves = new ArrayList<Turn>();
		firedMoves = new ArrayList<Turn>();
		nonFiredMoves = new ArrayList<Turn>();
		furtherMoves = new ArrayList<Turn>();

		/**
		 * This adds an empty turn to the list
		 * It is the turn without a move or a shot
		 */	
		//candidateMoves.add(newTurn);	

		/**
		 * This loop adds the turns where we attack but do not move
		 */	

		for (int i = 0; i < inputCopy.getWidth(); i++) {
			for (int j = 0; j < inputCopy.getHeight(); j++) {

				/**
				 * Quickly checks to see if there is a unit in the cell, if its on our team and if it's alive
				 */

				if (inputCopy.getUnitInNode(i,j) != null && inputCopy.getUnitInNode(i,j).getTeamID() == getCurrentPlayer().getPlayerID() && inputCopy.getUnitInNode(i, j).isAlive()) {
										
					/**
					 * Repeats the process but for enemy units
					 */

					for (int l = 0; l < inputCopy.getWidth(); l++) {
						for (int m = 0; m < inputCopy.getHeight(); m++) {
							if (inputCopy.getUnitInNode(l,m) != null && inputCopy.getUnitInNode(l,m).getTeamID() != getCurrentPlayer().getPlayerID()  && inputCopy.getUnitInNode(i, j).isAlive()){
								if (AttemptShot.attemptShot(i, j, l, m, inputCopy) == true) {
									
								
									Turn myTurn = new Turn(-1,-1,-1,-1, i, j, l, m, false, true, true);
									candidateMoves.add(myTurn);
									firedMoves.add(myTurn);
								}
							}
						}
					}
				}
			}
		}

		/**
		 * This loop covers where we both attack then move
		 */

		for (Turn t : firedMoves) {
			for (int i = 0; i < inputCopy.getWidth(); i++) {
				for (int j = 0; j < inputCopy.getHeight(); j++) {
					finder = new AStar(inputCopy.getMap(), inputCopy.getUnitInNode(t.shotSelectionX, t.shotSelectionY).getMoveDistance(), false);

					/**
					 * This checks that a path can be found between the two points
					 */

					if (finder.findPath(new UnitMover(inputCopy.getMap().getUnitInNode(t.shotSelectionX,t.shotSelectionY)), t.shotSelectionX, t.shotSelectionY, i,j) != null) {	
				
						Turn myTurn = new Turn(t.shotSelectionX,t.shotSelectionY, i, j, t.shotSelectionX, t.shotSelectionY, t.shotTargetX, t.shotTargetY, true, true, true);
						candidateMoves.add(myTurn);
					}
				}
			}
		}


		/**
		 * This loop covers where we move but do not fire
		 */

		for (int i = 0; i < inputCopy.getWidth(); i++) {
			for (int j = 0; j < inputCopy.getHeight(); j++) {
				
				if (inputCopy.getUnitInNode(i,j) != null && inputCopy.getUnitInNode(i,j).getTeamID() == getCurrentPlayer().getPlayerID()  && inputCopy.getUnitInNode(i, j).isAlive()) {
					
					for (int l = 0; l < inputCopy.getWidth(); l++) {
						for (int m = 0; m < inputCopy.getHeight(); m++) {
							finder = new AStar(inputCopy.getMap(), inputCopy.getUnitInNode(i,j).getMoveDistance(), false);
						
							Path foundPath = finder.findPath(new UnitMover(inputCopy.getMap().getUnitInNode(i,j)), i, j, l, m);
											
							if (foundPath != null) {
								//distancemoved
								int distanceMoved = foundPath.getLength();
								Turn myTurn = new Turn(i, j, l, m, -1, -1, -1, -1, true, false, false);	
								myTurn.distanceMoved = distanceMoved;
								candidateMoves.add(myTurn);
								nonFiredMoves.add(myTurn);
							}		
						}
					}
				}
			}
		}

		/**
		 * This loop covers where we move and then fire
		 */
		for (Turn t : nonFiredMoves) {
			for (int i = 0; i < inputCopy.getWidth(); i++) {
				for (int j = 0; j < inputCopy.getHeight(); j++) {
					//if theres a unit on the other team here
					if (inputCopy.getUnitInNode(i,j) != null && inputCopy.getUnitInNode(i,j).getTeamID() != getCurrentPlayer().getPlayerID()  && inputCopy.getUnitInNode(i, j).isAlive()) {

						//test the shot based on where the attacking unit would be now
						if (AttemptShot.testShot(t.moveTargetX, t.moveTargetY, i, j, inputCopy, inputCopy.getUnitInNode(i, j).getShootDistance(), t.moveSelectionX, t.moveSelectionY) == true) {	
						
							Turn myTurn = new Turn(t.moveSelectionX, t.moveSelectionY, t.moveTargetX, t.moveTargetY, t.moveTargetX, t.moveTargetY, i, j, true, true, false);							
							myTurn.distanceMoved = t.distanceMoved;
							candidateMoves.add(myTurn);		
							furtherMoves.add(myTurn);
						}
					}
				}		
			}
		}
		//now for the moves where we continue after attacking
		for (Turn t : furtherMoves) {
			//the original move distance of the unit, found by looking at the unit we originally moved
			int originalDist = inputCopy.getUnitInNode(t.moveSelectionX, t.moveSelectionY).getMoveDistance();
			//a new distance calculated from the distance moved on the previous turn
			int newDist = originalDist - t.distanceMoved;
			if (newDist > 0){
				//System.out.println("doing this loop!");
				for (int i = 0; i < inputCopy.getWidth(); i++) {
					for (int j = 0; j < inputCopy.getHeight(); j++) {
						//for each cell on the map, create a new pathfinder that can search as far as the remaining distance
						finder = new AStar(inputCopy.getMap(), newDist, false);
						//we use the unit found at the original move's location as a mover, the shot selection from the turn as the start point, and the iteration values as the end
						Path foundPath = finder.findPath(new UnitMover(inputCopy.getMap().getUnitInNode(t.moveSelectionX,t.moveSelectionY)), t.shotSelectionX, t.shotSelectionY, i, j);
						if (foundPath != null) {
							
							Turn myTurn = new Turn(t.moveSelectionX, t.moveSelectionY, t.moveTargetX, t.moveTargetY, t.shotSelectionX, t.shotSelectionY, t.shotTargetX, t.shotTargetY, i, j, true, true, false, true);	
							candidateMoves.add(myTurn);
						}								
					}
				}
			}
		}
		
		/**
		 * This loop looks at how much further the unit can move after firing
		 */
		
		Collections.sort(candidateMoves);
		int n = 24;
		if (candidateMoves.size() < 24) { n = candidateMoves.size(); }		
		if (largeCull == true) { n = 10; }
		List<Turn> sublist = new ArrayList<Turn>(candidateMoves.subList(0,n));
		//return sublist;
		return sublist;
	}

	/**
	 * This class is essentially the same as the previous class, except that it builds turns based on the opposing player's units
	 * @param input: this is the input game state
	 */

	public static List<Turn> getLegalMovesMin(GameMap input, boolean largeCull){

		GameMap inputCopy = new GameMap(input); 	
		Turn newTurn = new Turn(-1,-1,-1,-1,-1,-1,-1,-1,false,false,false);
		List<Turn> candidateMoves;
		List<Turn> firedMoves;
		List<Turn> nonFiredMoves;
		List<Turn> furtherMoves;
		candidateMoves = new ArrayList<Turn>();
		firedMoves = new ArrayList<Turn>();
		nonFiredMoves = new ArrayList<Turn>();
		furtherMoves = new ArrayList<Turn>();

		//candidateMoves.add(newTurn);	

		for (int i = 0; i < inputCopy.getWidth(); i++) {
			for (int j = 0; j < inputCopy.getHeight(); j++) {

				if (inputCopy.getUnitInNode(i,j) != null && inputCopy.getUnitInNode(i,j).getTeamID() != getCurrentPlayer().getPlayerID() && inputCopy.getUnitInNode(i, j).isAlive()) {

					for (int l = 0; l < inputCopy.getWidth(); l++) {
						for (int m = 0; m < inputCopy.getHeight(); m++) {
							if (inputCopy.getUnitInNode(l,m) != null && inputCopy.getUnitInNode(l,m).getTeamID() == getCurrentPlayer().getPlayerID()  && inputCopy.getUnitInNode(i, j).isAlive()){
								if (AttemptShot.attemptShot(i, j, l, m, inputCopy) == true) {

									Turn myTurn = new Turn(-1,-1,-1,-1, i, j, l, m, false, true, true);
									candidateMoves.add(myTurn);
									firedMoves.add(myTurn);
								}
							}
						}
					}
				}
			}
		}

		for (Turn t : firedMoves) {
			for (int i = 0; i < inputCopy.getWidth(); i++) {
				for (int j = 0; j < inputCopy.getHeight(); j++) {
					finder = new AStar(inputCopy.getMap(), inputCopy.getUnitInNode(t.shotSelectionX, t.shotSelectionY).getMoveDistance(), false);
					if (finder.findPath(new UnitMover(inputCopy.getMap().getUnitInNode(t.shotSelectionX,t.shotSelectionY)), t.shotSelectionX, t.shotSelectionY, i,j) != null) {				
						Turn myTurn = new Turn(t.shotSelectionX,t.shotSelectionY, i, j, t.shotSelectionX, t.shotSelectionY, t.shotTargetX, t.shotTargetY, true, true, true);
						candidateMoves.add(myTurn);
					}
				}
			}
		}
		
		

		for (int i = 0; i < inputCopy.getWidth(); i++) {
			for (int j = 0; j < inputCopy.getHeight(); j++) {
				if (inputCopy.getUnitInNode(i,j) != null && inputCopy.getUnitInNode(i,j).getTeamID() != getCurrentPlayer().getPlayerID()  && inputCopy.getUnitInNode(i, j).isAlive()) {	

					for (int l = 0; l < inputCopy.getWidth(); l++) {
						for (int m = 0; m < inputCopy.getHeight(); m++) {
							finder = new AStar(inputCopy.getMap(), inputCopy.getUnitInNode(i,j).getMoveDistance(), false);						
							Path foundPath = finder.findPath(new UnitMover(inputCopy.getMap().getUnitInNode(i,j)), i, j, l, m);	
						
							if (foundPath != null) {	
								int distanceMoved = foundPath.getLength();
								Turn myTurn = new Turn(i, j, l, m, -1, -1, -1, -1, true, false, false);	
								
								myTurn.distanceMoved = distanceMoved;
								candidateMoves.add(myTurn);
								nonFiredMoves.add(myTurn);
							}		
						}
					}
				}
			}
		}

		for (Turn t : nonFiredMoves) {
			for (int i = 0; i < inputCopy.getWidth(); i++) {
				for (int j = 0; j < inputCopy.getHeight(); j++) {
					if (inputCopy.getUnitInNode(i,j) != null && inputCopy.getUnitInNode(i,j).getTeamID() == getCurrentPlayer().getPlayerID()  && inputCopy.getUnitInNode(i, j).isAlive()) {
						if (AttemptShot.testShot(t.moveTargetX, t.moveTargetY, i, j, inputCopy, inputCopy.getUnitInNode(i, j).getShootDistance(), t.moveSelectionX, t.moveSelectionY) == true) {						
							Turn myTurn = new Turn(t.moveSelectionX, t.moveSelectionY, t.moveTargetX, t.moveTargetY, t.moveTargetX, t.moveTargetY, i, j, true, true, false);	
							myTurn.distanceMoved = t.distanceMoved;
							candidateMoves.add(myTurn);		
							furtherMoves.add(myTurn);
						}
					}
				}		
			}
		}
		
		for (Turn t : furtherMoves) {
			//the original move distance of the unit, found by looking at the unit we originally moved
			int originalDist = inputCopy.getUnitInNode(t.moveSelectionX, t.moveSelectionY).getMoveDistance();
			//a new distance calculated from the distance moved on the previous turn
			int newDist = originalDist - t.distanceMoved;
			if (newDist > 0){
				//System.out.println("doing this loop!");
				for (int i = 0; i < inputCopy.getWidth(); i++) {
					for (int j = 0; j < inputCopy.getHeight(); j++) {
						//for each cell on the map, create a new pathfinder that can search as far as the remaining distance
						finder = new AStar(inputCopy.getMap(), newDist, false);
						//we use the unit found at the original move's location as a mover, the shot selection from the turn as the start point, and the iteration values as the end
						Path foundPath = finder.findPath(new UnitMover(inputCopy.getMap().getUnitInNode(t.moveSelectionX,t.moveSelectionY)), t.shotSelectionX, t.shotSelectionY, i, j);
						if (foundPath != null) {
							
							Turn myTurn = new Turn(t.moveSelectionX, t.moveSelectionY, t.moveTargetX, t.moveTargetY, t.shotSelectionX, t.shotSelectionY, t.shotTargetX, t.shotTargetY, i, j, true, true, false, true);	
							candidateMoves.add(myTurn);
						}								
					}
				}
			}
		}
		
		Collections.sort(candidateMoves);
		int n = 24;
		if (candidateMoves.size() < 24) { n = candidateMoves.size(); }
		if (largeCull == true) { n = 10; }
		List<Turn> sublist = new ArrayList<Turn>(candidateMoves.subList(0,n));
		//return sublist;
		return sublist;
	}

	/**
	 * This class builds a new state based on a turn that is fed to it
	 * @param state: the state from which a new state is built from
	 * @param m: the turn which contains the values to build a new state
	 * @return: a new state
	 */
	
	public static GameMap getNextState(GameMap state, Turn m) {
		GameMap newState = new GameMap(state); 

		/**
		 * If the turn has a unit fire, and that unit fires before moving
		 */
		if (m.firedFirst == true && m.turnFired == true) {	
			
			/**
			 * Kill the target unit
			 */
			
			newState.getUnitInNode(m.shotTargetX, m.shotTargetY).setAlive();						
			
			/** 
			 * If the unit also moves, remove it from the old cell and place it in the new one
			 */
			
			if (m.turnMoved == true) {
				units.UnitInterface myUnit = newState.getUnitInNode(m.moveSelectionX, m.moveSelectionY); 			
				newState.clearUnitInCell(m.moveSelectionX, m.moveSelectionY);			
				newState.setUnitInCell(m.moveTargetX, m.moveTargetY, myUnit);	
			}			
			//System.out.println("State generated");
			return newState;
		}
		
		/**
		 * If the turn does not contain a shot first, then move the unit first instead
		 */

		else if (m.firedFirst == false && m.turnMoved == true) {

			units.UnitInterface myUnit= newState.getUnitInNode(m.moveSelectionX, m.moveSelectionY);			
			newState.clearUnitInCell(m.moveSelectionX, m.moveSelectionY);				
			newState.setUnitInCell(m.moveTargetX, m.moveTargetY, myUnit);		

			/**
			 * Now if the move contains an attack, kill the targeted unit
			 */
			
			if (m.turnFired == true) {
				newState.getUnitInNode(m.shotTargetX, m.shotTargetY).setAlive();	
			}		
			
			/**
			 * Now if we move again
			 */
			
			if (m.extraMove == true) {
				units.UnitInterface myUnit2 = newState.getUnitInNode(m.moveTargetX, m.moveTargetY);	
				newState.setUnitInCell(m.newTargetX, m.newTargetY, myUnit2);
				newState.clearUnitInCell(m.moveTargetX, m.moveTargetY);
				
			}
			//System.out.println("State generated");
			return newState;
		}

		/**
		 * If the turn contains neither a move or a shot, simply return the old state
		 */
		
		else if (m.turnMoved == false && m.turnFired == false){
			System.out.println("State generated");
			return newState;
		}

		return null;
	}
	
	/**
	 * This is similar to the previous method except it generates the next state that will be fed to the game itself
	 * Hence it does not need to be given an input state
	 */

	public void doAIMove(Turn m, GameMap map) {

		if (m.turnMoved == false && m.turnFired == false) {
			/**
			 * Nothing happens if there is no move or shot made
			 */
		}

		else if (m.firedFirst == true && m.turnFired == true) {		
			try {
				Thread.sleep(500);
				} catch(InterruptedException e) {
				} 
			getMap().getUnitInNode(m.shotTargetX, m.shotTargetY).setAlive();
			getCurrentPlayer().setFired(true);
			if (getCurrentPlayer().getPlayerID() == 0) { map.blueUnitCount = map.blueUnitCount -1; }
			else if (getCurrentPlayer().getPlayerID() == 1) { map.redUnitCount = map.redUnitCount -1; }
			repaint(0);


			if (m.turnMoved == true) {
				try {
					Thread.sleep(500);
					} catch(InterruptedException e) {
					} 
				units.UnitInterface myUnit = getMap().getUnitInNode(m.moveSelectionX, m.moveSelectionY);
				getMap().setUnitInCell(m.moveTargetX, m.moveTargetY, myUnit);
				getMap().clearUnitInCell(m.moveSelectionX, m.moveSelectionY);
				myUnit = null;
				getCurrentPlayer().setMoved(true);
			}
			
			
			
		}

		else if (m.firedFirst == false && m.turnMoved == true) {

			try {
				Thread.sleep(500);
				} catch(InterruptedException e) {
				} 
			units.UnitInterface myUnit = getMap().getUnitInNode(m.moveSelectionX, m.moveSelectionY);
			getMap().setUnitInCell(m.moveTargetX, m.moveTargetY, myUnit);
			getMap().clearUnitInCell(m.moveSelectionX, m.moveSelectionY);
			myUnit = null;
			getCurrentPlayer().setMoved(true);
			repaint(0);
			

			if (m.turnFired == true) {		
				try {
					Thread.sleep(500);
					} catch(InterruptedException e) {
					} 
				getMap().getUnitInNode(m.shotTargetX, m.shotTargetY).setAlive();
				getCurrentPlayer().setFired(true);
				if (getCurrentPlayer().getPlayerID() == 0) { map.blueUnitCount = map.blueUnitCount -1; }
				else if (getCurrentPlayer().getPlayerID() == 1) { map.redUnitCount = map.redUnitCount -1; }
				repaint(0);
				
			}
			
			/**
			 * Now if we move again
			 */
			
			if (m.extraMove == true) {
				try {
					Thread.sleep(500);
					} catch(InterruptedException e) {
					} 
				units.UnitInterface myUnit2 = getMap().getUnitInNode(m.moveTargetX, m.moveTargetY);		
				getMap().setUnitInCell(m.newTargetX, m.newTargetY, myUnit2);
				getMap().clearUnitInCell(m.moveTargetX, m.moveTargetY);
				
			
			}
		}
	}
	
	/**
	 * Check if a state is a final state
	 * @param s: an input state
	 * @return either whether the state is a winning state or a losing state
	 */

	public static boolean isFinal(GameMap s) {
		return isWinning(s) || isLosing(s);
	}

	/**
	 * Checks to see if there the state is a losing state for an AI player
	 */

	public static boolean isLosing(GameMap s) {
		if (getCurrentPlayer().getPlayerID() == 0) {
			if (s.redUnitCount == 0) {
				return true;
			}
		}

		if (getCurrentPlayer().getPlayerID() == 1) {
			if (s.blueUnitCount == 0) {
				return true;
			}
		}
		return false; 
	}
	
	/**
	 * Checks to see if there the state is a winning state for an AI player
	 */

	public static boolean isWinning(GameMap s) {
		if (getCurrentPlayer().getPlayerID() == 0) {
			if (s.blueUnitCount == 0) {
				System.out.println("no blues");
				return true;
			}
		}

		if (getCurrentPlayer().getPlayerID() == 1) {
			if (s.redUnitCount == 0) {
				System.out.println("no reds");
				return true;
			}
		}
		return false;
	}

	/**
	 * This class handles what happens when the player clicks their mouse
	 * @param x: the x co-ordinate of the mouse click
	 * @param y: the y co-ordinate of the mouse click
	 * @param player: the player object that clicked
	 */

	void handleMousePressed(int x, int y, PlayerInterface player, GameMap m) {
		
		/**
		 * The use of a layeredpane and border in the graphics package object creates an offset to the x/y values
		 * First we subtract these offsets to get the x/y co-ordinates on the playing field
		 * We divide these co-ordinates by 32: each cell image is 32x32, so doing this gives us the exact co-ordinate of the cell the player clicked
		 */

		x = x - 63;
		y = y - 84;
		x = x / 32;
		y = y / 32;	

		if ((x < 0) || (y < 0) || (x >= getMap().getWidth()) || (y >= getMap().getHeight())) {
			return;
		}

		/**
		 * If there is a unit in the tile
		 * If it is on the opposing team
		 * If there is already something selected and the player has not yet fired
		 * Check if the shot is legal and if it is, kill the unit and lower the unit counter
		 */

		if (getMap().getUnitInNode(x, y) != null) {
			if (getMap().getUnitInNode(x, y).getTeamID() != getTurnID()) {
				if (player.getSelectedx() != -1 && player.getSelectedy() !=-1) {
					if (player.isFired() == false) {

						System.out.println("Variables fed to attempt shot: sx =" + player.getSelectedx() + ", sy =" + player.getSelectedy() + ", tx =" + x + ", ty = " + y);
						if (AttemptShot.attemptShot(player.getSelectedx(), player.getSelectedy(), x, y, getMap()) == true){

							if (getMap().getUnitInNode(x, y).getTeamID() == 0) {  m.redUnitCount = m.redUnitCount -1; }
							else if (getMap().getUnitInNode(x, y).getTeamID() == 1) { m.blueUnitCount = m.blueUnitCount -1;}

							getMap().getUnitInNode(x, y).setAlive();
							repaint(0);
							player.setFired(true);
						}
					}	
				}		
			} 
			
			/**
			 * If the unit is on our team, select it and grab the movedistance
			 */
			
			else {
				if (player.isMovedOnce() == false){
				player.setSelectedx(x);
				player.setSelectedy(y);
				player.setLastFindx(- 1);
				
				int movedist = getMap().getUnitInNode(x, y).getMoveDistance();
				player.setMoverange(movedist);}
			}	
			
		/**
		 * Otherwise if the player hasn't already moved and there isn't a unit where we clicked
		 * Check to see if we can legally move the unit there, and if we can, do so
		 */
			
		} else if (player.isMoved() == false) {

			if (player.getSelectedx() != -1) {
				getMap().clearNodeVisited();
				path = finder.findPath(new UnitMover(getMap().getUnitInNode(player.getSelectedx(),  player.getSelectedy())), player.getSelectedx(), player.getSelectedy(), x, y);
				int pathLength = finder.findPath(new UnitMover(getMap().getUnitInNode(player.getSelectedx(),  player.getSelectedy())), player.getSelectedx(), player.getSelectedy(), x, y).getLength();

				/**
				 * This checks to see if it is the first time a unit has been moved on a turn
				 */
				
				if (player.isMovedOnce() == false){
					player.setMoveDistanceTracker(getMap().getUnitInNode(player.getSelectedx(),  player.getSelectedy()).getMoveDistance());
				}

				if (path != null) {
					path = null;

					UnitInterface unit = getMap().getUnitInNode(player.getSelectedx(), player.getSelectedy());
					getMap().clearUnitInCell(player.getSelectedx(), player.getSelectedy());
					getMap().setUnitInCell(x, y, unit);

					/**
					 * Subtract the distance moved from the tracker so we know how many more space we can still move this turn
					 * Then update the moverange
					 */
					player.setMoveDistanceTracker(player.getMoveDistanceTracker() - pathLength + 1);
					player.setMoverange(player.getMoveDistanceTracker());
					player.setMovedOnce(true);

					/**
					 * If we've used up a players move distance, we can set moved = true
					 */
					if (player.getMoverange() == 0) {
						player.setMoved(true);
						player.setMovedOnce(false);
					}
					
					player.setSelectedx(x);
					player.setSelectedy(y);
					System.out.println("Unit moved to " + player.getSelectedx() + ", " + player.getSelectedy());
					player.setLastFindx(- 1);
				}
			}
		}
		
		/**
		 * If a player has already moved/fired on a turn, switch the turn to the opposing player by calling the swap player class
		 */
		
		if (player.isMoved() == true && player.isFired() == true) {
			repaint();
			player.swapPlayer(this);
		}
		System.out.print(player.isMoved());
		System.out.print(" " + player.isFired());
		repaint(0);

		if (m.redUnitCount == 0) { System.out.println("Blue Team Won the Game!");}
		else if (m.blueUnitCount == 0) { System.out.println ("Red Team Won the Game");}
	}

	/**
	 * This method is called whenever the player moves their mouse
	 */
	
	void handleMouseMoved(int x, int y, PlayerInterface player) {
		x = x - 63;
		y = y - 84;
		x = x / 32;
		y = y / 32;

		if ((x < 0) || (y < 0) || (x >= getMap().getWidth()) || (y >= getMap().getHeight())) {
			return;
		}

		if (player.getSelectedx() != -1) {
			if ((player.getLastFindx() != x) || (player.getLastFindY() != y)) {
				player.setLastFindx(x);
				player.setLastFindY(y);
				path = finder.findPath(new UnitMover(getMap().getUnitInNode(player.getSelectedx(),  player.getSelectedy())), player.getSelectedx(), player.getSelectedy(), x, y);
				repaint(0);
			}
		}
	}

	/**
	 * Miscellaneous getters and setters
	 */

	public GameMap getMap() {
		return map;
	}

	public void setMap(GameMap map) {
		this.map = map;
	}

	public Path getPath() {
		return path;
	}

	public void setPath(Path path) {
		this.path = path;
	}

	public int getTurnID() {
		return turnID;
	}

	public void setTurnID(int turnID) {
		this.turnID = turnID;
	}


	public players.PlayerInterface getPlayer2() {
		return player2;
	}


	public players.PlayerInterface getPlayer1() {
		return player1;
	}


	public static players.PlayerInterface getCurrentPlayer() {
		return currentPlayer;
	}


	public static void setCurrentPlayer(players.PlayerInterface currentPlayer) {
		Game.currentPlayer = currentPlayer;
	}

	public Game instanceRetriever(){
		return this;
	}
	
	

}
