package game;

import java.io.Serializable;
import java.util.List;
import java.util.ArrayList;

import pathfinder.Mover;
import game.MapLayouts;
import game.MapLayouts.unitInitialiser;
import units.*;


public class GameMap implements pathfinder.MapInterface, Serializable {

	private static final long serialVersionUID = 2793386624586663925L;
	
	/**
	 * These variables hold map height and width for the map instance
	 */
	public int MAP_WIDTH;
	public int MAP_HEIGHT;

	/**
	 * These variables specify the integer value for each tile type/unit
	 * They can be interpreted by various classes to determine what tile to draw,
	 * or whether a shot is legal, etc
	 */
	public static final int GRASS = 0;
	public static final int TREE = 1;
	public static final int WATER = 2;
	public static final int WALL = 3;
	public static final int SOLDIER_RED = 4;
	public static final int SOLDIER_BLUE = 5;
	public static final int COMM_RED = 6;
	public static final int COMM_BLUE = 7;
	public static final int SNIPER_RED = 8;
	public static final int SNIPER_BLUE = 9;
	public static final int GREN_RED = 10;
	public static final int GREN_BLUE = 11;

	/**
	 * This stores the list of units present in the map
	 */	
	public List<UnitInterface> builtUnits = new ArrayList<UnitInterface>();

	public int redUnitCount = 0; 
	public int blueUnitCount = 0;

	/**
	 * This variable is given to maplayouts so it knows what map to generate
	 */
	
	public static int mapselection = 2;

	/**
	 * This is the grid of tiles
	 */
	public int[][] tiletypes;


	/**
	 * This is the grid of units and map of visited nodes for the pathfinder
	 */
	
	public UnitNode[][] unitpositions;
	private boolean[][] nodevisited;

	public static boolean[][] legalShots; 
	public static boolean[][] legalMoves;
	
	/**
	 * Used by the AI to determine which Turn object spawned this state
	 */

	public Turn turnUsed = null;

	/**
	 * Creates an instance of this class, using mapselection as a template
	 */
	
	public GameMap() {
		
		MapLayouts.createMap(mapselection, this);
		List<unitInitialiser> unitsToBuild = MapLayouts.createUnits(mapselection);

		for (unitInitialiser u: unitsToBuild){
			if (u.type == 1){
				UnitInterface newUnit = new units.RedSoldier(u);
				this.builtUnits.add(newUnit);
			}

			if (u.type == 2){
				UnitInterface newUnit = new units.RedCommando(u);
				this.builtUnits.add(newUnit);
			}

			if (u.type == 3){
				UnitInterface newUnit = new units.RedSniper(u);
				this.builtUnits.add(newUnit);
			}

			if (u.type == 4){
				UnitInterface newUnit = new units.BlueSoldier(u);
				this.builtUnits.add(newUnit);
			}

			if (u.type == 5){
				UnitInterface newUnit = new units.BlueCommando(u);
				this.builtUnits.add(newUnit);
			}

			if (u.type == 6){
				UnitInterface newUnit = new units.BlueSniper(u);
				this.builtUnits.add(newUnit);
			}	
			
			if (u.type == 7){
				UnitInterface newUnit = new units.RedGrenadier(u);
				this.builtUnits.add(newUnit);
			}	
			if (u.type == 8){
				UnitInterface newUnit = new units.BlueGrenadier(u);
				this.builtUnits.add(newUnit);
			}	
		}

		unitpositions = new UnitNode[MAP_WIDTH][MAP_HEIGHT];
		fillUnits(MAP_WIDTH, MAP_HEIGHT);
		nodevisited = new boolean[MAP_WIDTH][MAP_HEIGHT];
		
		/**
		 * BuiltUnits was filled based on the unitInitialisers we found
		 * We loop through the list and use its startposition variables 
		 * to choose where to place it
		 */

		for (UnitInterface u: builtUnits){
			unitpositions[u.getStartpositionx()][u.getStartpositiony()].unitInCell = u;
			if (u.getTeamID() == 0) { redUnitCount += 1;}
			else if (u.getTeamID() == 1) { blueUnitCount += 1;}
		}
	}

	/**
	 * This is a copy constructor that is used to clone an input map along with
	 * all of its instance variables, rather than just referencing an existing one
	 * @param old: an old gamemap to copy
	 */

	public GameMap(GameMap old) {

		this();	
		this.MAP_WIDTH = old.getWidth();
		this.MAP_HEIGHT = old.getHeight();
		this.builtUnits = new ArrayList<UnitInterface>();
		this.tiletypes = old.tiletypes;
		this.nodevisited = old.nodevisited;
		this.unitpositions =  new UnitNode[this.MAP_WIDTH][this.MAP_HEIGHT];
		

		this.redUnitCount = old.redUnitCount;
		this.blueUnitCount = old.blueUnitCount;

		/**
		 * Create a new list of unit instances based off clones of old ones
		 */
		
		for (UnitInterface u: old.builtUnits){

			if (u.getimageloc() == 4){
				UnitInterface copyUnit1 = new RedSoldier(u);
				this.builtUnits.add(copyUnit1);
			}
			if (u.getimageloc() == 5){
				UnitInterface copyUnit2 = new BlueSoldier(u);
				this.builtUnits.add(copyUnit2);
			}
			if (u.getimageloc() == 6){
				UnitInterface copyUnit3 = new RedCommando(u);
				this.builtUnits.add(copyUnit3);
			}
			if (u.getimageloc() == 7){
				UnitInterface copyUnit4 = new BlueCommando(u);
				this.builtUnits.add(copyUnit4);
			}
			if (u.getimageloc() == 8){
				UnitInterface copyUnit5 = new RedSniper(u);
				this.builtUnits.add(copyUnit5);
			}
			if (u.getimageloc() == 9){
				UnitInterface copyUnit6 = new BlueSniper(u);
				this.builtUnits.add(copyUnit6);
			}
			if (u.getimageloc() == 10){
				UnitInterface copyUnit7 = new RedGrenadier(u);
				this.builtUnits.add(copyUnit7);
			}
			if (u.getimageloc() == 11){
				UnitInterface copyUnit8 = new BlueGrenadier(u);
				this.builtUnits.add(copyUnit8);
			}
		}

		/**
		 * Copy the grid of unitNodes
		 */
		for(int i=0;i<old.unitpositions.length;i++){
			for(int j=0;j<old.unitpositions[i].length;j++){
				this.unitpositions[i][j]=new UnitNode(i,j);
			}
		}

		/**
		 * Update the grid of unitNodes so it points to our new units
		 */
		for (UnitInterface u: this.builtUnits){
		}
		
		
		for (UnitInterface u: this.builtUnits){
			this.unitpositions[u.getCurrentPosX()][u.getCurrentPosY()].unitInCell = u;	
			if (u.isAlive() == true) {
			}
		}


		for (int i=0; i<nodevisited.length;i++){
			for (int j=0; j<nodevisited[i].length;j++){
				nodevisited[i][j]=old.nodevisited[i][j];
			}
		}
		
		
	}


	public void clearNodeVisited() {
		for (int x=0; x < getWidth(); x++) {
			for (int y=0; y < getHeight(); y++) {
				nodevisited[x][y] = false;
			}
		}
	}

	public void fillUnits(int width, int height) {
		for (int xpos=0; xpos < width; xpos++) {
			for (int ypos=0; ypos < height; ypos++) {
				unitpositions[xpos][ypos] = new UnitNode(xpos,ypos);			
			}
		}
	}

	/**
	 * Class used to build the map of unit locations
	 * @param x: the x co-ordinate of the tile
	 * @param y: the y co-ordinate of the tile
	 * @param unitInCell: a pointer to a unit instance that is currently in the cell
	 */
	public class UnitNode implements Serializable {

		private static final long serialVersionUID = -1409473536405419428L;
		public int x;
		public int y;
		public UnitInterface unitInCell = null;

		public UnitNode(int x, int y) {
			this.x = x;
			this.y = y;
		}

		public UnitNode(UnitNode un) {
			this.x = un.x;
			this.y = un.y;
			unitInCell=un.unitInCell;
		}
	}

	public GameMap getMap(){
		return this;
	}

	public UnitNode getNode(int x, int y) {
		return unitpositions[x][y];
	}


	public boolean nodevisited(int x, int y) {
		return nodevisited[x][y];
	}

	public int getTile(int x, int y) {
		return tiletypes[x][y];
	}

	public UnitInterface getUnitInNode(int x, int y) {
		return this.unitpositions[x][y].unitInCell;
	}

	public void clearUnitInCell(int x, int y) {
		unitpositions[x][y].unitInCell = null;
	}

	public void setUnitInCell(int x, int y, UnitInterface unit) {
		unitpositions[x][y].unitInCell = unit;
	}


	public boolean pathBlocked(Mover mover, int x, int y) {

		if (getUnitInNode(x,y) != null && getUnitInNode(x,y).isAlive() == true) {
			return true;
		}

		if (getUnitInNode(x,y) != null && getUnitInNode(x,y).isAlive() == false) {
			return false;
		}

		if (getTile(x,y) == 2) {
			return true;
		}

		if (getTile(x,y) == 3) {
			return true;
		}		

		return false;
	}

	public float getCost(Mover mover, int sx, int sy, int tx, int ty) {
		return 1;
	}

	public int getHeight() {
		return MAP_HEIGHT;
	}

	public int getWidth() {
		return MAP_WIDTH;
	}

	public void pathNodeVisited(int x, int y) {
		nodevisited[x][y] = true;
	}
}


