package game;

import game.GameInterface.MapSelect;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.EventQueue;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;

import javax.imageio.ImageIO;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.border.EmptyBorder;
import javax.swing.JLayeredPane;
import javax.swing.JButton;
import javax.swing.JLabel;
import java.awt.GridLayout;
import javax.swing.JComboBox;
import javax.swing.DefaultComboBoxModel;

import units.UnitInterface;

public class MapEditor extends JFrame {

	/**
	 * This class implements the Map Editor functionality. It allows players to modify
	 * map layouts
	 */

	//Store the image tiles for rendering
	private Image[] tiles = new Image[13];
	private Image buffer;

	//Creates a GameMap instance
	private GameMap editor = new GameMap();

	public static GameMap savedMap = new GameMap();

	//Creates the JPanel within the JFrame
	private JPanel contentPane;

	public MapEditor() {
		//Set Window Options
		setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		int windowX = ((editor.getWidth() * 32) + 100);
		int windowY = ((editor.getHeight() * 32) + 150);
		setBounds(100, 100, 657, 644);
		contentPane = new JPanel();
		contentPane.setBorder(new EmptyBorder(5, 5, 5, 5));
		contentPane.setLayout(new BorderLayout(0, 0));
		setContentPane(contentPane);
		setSize(windowX,windowY);
		setResizable(true);
		setVisible(true);

		//Creates the layered pane for drawing and buttons
		JLayeredPane layeredPane = new JLayeredPane();
		contentPane.add(layeredPane, BorderLayout.CENTER);

		//Sets up the graphics panel
		gameGraphics gGraphics = new gameGraphics();
		gGraphics.setSize(1024,1024);
		gGraphics.setOpaque(false);

		//Adds the graphics panel at the lowest layer
		layeredPane.add(gGraphics, new Integer(0)); 

		//Buttons...
		JButton btnAddUnits = new JButton("Change Units");
		layeredPane.setLayer(btnAddUnits, 1);
		btnAddUnits.setBounds(144, 11, 110, 23);
		btnAddUnits.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				ChoseUnitWindow unitwindow = new ChoseUnitWindow(editor, instanceRetriever());
				unitwindow.setVisible(true);
			}
		});
		
		final JComboBox cB_Slot = new JComboBox();
		layeredPane.setLayer(cB_Slot, 1);
		cB_Slot.setModel(new DefaultComboBoxModel(new String[] {"Slot 1", "Slot 2", "Slot 3", "Slot 4", "Slot 5"}));
		cB_Slot.setBounds(144, 45, 85, 20);
		layeredPane.add(cB_Slot);
		layeredPane.add(btnAddUnits);

		JButton btnExit = new JButton("Exit");
		layeredPane.setLayer(btnExit, 1);
		btnExit.setBounds(264, 11, 61, 23);
		btnExit.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				instanceRetriever().dispose();
			}
		});
		layeredPane.add(btnExit);
		
		/**
		 * Sets up a map for being cloned and transfers it into savedMap for use by
		 * the game or serialiser
		 */

		JButton btnSave = new JButton("Save Layout");
		layeredPane.setLayer(btnSave, 1);
		btnSave.setBounds(10, 11, 124, 23);
		btnSave.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				
				/**
				 * Fill builtUnits with all the units we may have created
				 */
				
				editor.builtUnits = new ArrayList<units.UnitInterface>();
				for (int i = 0; i < editor.getWidth(); i++){
					for (int j = 0; j < editor.getHeight(); j++){
						if (editor.getUnitInNode(i, j) != null){
							editor.getUnitInNode(i, j).setCurrentPosX(i);
							editor.getUnitInNode(i, j).setCurrentPosY(j);
							editor.builtUnits.add(editor.getUnitInNode(i, j));
							
						}
					}
				}
				
				/**
				 * Count how many units are present
				 */
				
				editor.blueUnitCount = 0;
				editor.redUnitCount = 0;
				for (UnitInterface u : editor.builtUnits){
					if (u.getTeamID() == 0){ editor.redUnitCount +=1;}
					else if (u.getTeamID()==1){editor.blueUnitCount+=1;}
				}
				
				/**
				 * Clone editor into savedMap
				 */
				
				savedMap = new GameMap(editor);
				GameInterface.mapSaved = true;
			}
		});
		layeredPane.add(btnSave);
		
		/**
		 * Similar functionality to the above button but serialises the file as well
		 */
		
		JButton btnSaveToFile = new JButton("Save to File");
		layeredPane.setLayer(btnSaveToFile, 1);
		btnSaveToFile.setBounds(10, 45, 117, 23);
		btnSaveToFile.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				
				editor.builtUnits = new ArrayList<units.UnitInterface>();
				for (int i = 0; i < editor.getWidth(); i++){
					for (int j = 0; j < editor.getHeight(); j++){
						if (editor.getUnitInNode(i, j) != null){
							editor.getUnitInNode(i, j).setCurrentPosX(i);
							editor.getUnitInNode(i, j).setCurrentPosY(j);
							editor.builtUnits.add(editor.getUnitInNode(i, j));
							
						}
					}
				}
				
				editor.blueUnitCount = 0;
				editor.redUnitCount = 0;
				for (UnitInterface u : editor.builtUnits){
					if (u.getTeamID() == 0){ editor.redUnitCount +=1;}
					else if (u.getTeamID()==1){editor.blueUnitCount+=1;}
				}
				
				savedMap = new GameMap(editor);				
				System.out.println("blue " + savedMap.blueUnitCount);
				
				/**
				 * Create a serializer and save the map to file
				 */
				
				MapSerializer serial = new MapSerializer();
				int saveslot = cB_Slot.getSelectedIndex();
				serial.SaveMap(savedMap, saveslot);
			}
		});
		layeredPane.add(btnSaveToFile);

		try {
			tiles[GameMap.GRASS] = ImageIO.read(getResource("res/tile/GRASS.png"));
			tiles[GameMap.TREE] = ImageIO.read(getResource("res/tile/TREE.png"));
			tiles[GameMap.WATER] = ImageIO.read(getResource("res/tile/WATER.png"));
			tiles[GameMap.WALL] = ImageIO.read(getResource("res/tile/WALL.png"));
			tiles[GameMap.SOLDIER_RED] = ImageIO.read(getResource("res/unit/SOLDIER_RED.png"));
			tiles[GameMap.SOLDIER_BLUE] = ImageIO.read(getResource("res/unit/SOLDIER_BLUE.png"));
			tiles[GameMap.COMM_RED] = ImageIO.read(getResource("res/unit/COMM_RED.png"));
			tiles[GameMap.COMM_BLUE] = ImageIO.read(getResource("res/unit/COMM_BLUE.png"));
			tiles[GameMap.SNIPER_RED] = ImageIO.read(getResource("res/unit/SNIPER_RED.png"));
			tiles[GameMap.SNIPER_BLUE] = ImageIO.read(getResource("res/unit/SNIPER_BLUE.png"));	
			tiles[GameMap.GREN_RED] = ImageIO.read(getResource("res/unit/GRENADIER_RED.png"));
			tiles[GameMap.GREN_BLUE] = ImageIO.read(getResource("res/unit/GRENADIER_BLUE.png"));	
		}

		catch (IOException e) {
			System.err.println("Failed to load resources: "+e.getMessage());
			System.exit(0);
		}

		addMouseListener(new MouseAdapter(){
			public void mousePressed(MouseEvent e) {
				handleMousePressed(e.getX(), e.getY());		
			}
		});	
	}
	
	public MapEditor instanceRetriever(){
		return this;
	}

	private InputStream getResource(String ref) throws IOException {
		return new FileInputStream(ref);
	}

	class gameGraphics extends JPanel {

		public void paint(Graphics graphics) {	
			
			System.out.println("Map width " + editor.MAP_WIDTH);
			System.out.println("Map height " + editor.MAP_HEIGHT);

			if (buffer == null) {
				buffer = new BufferedImage(1024, 1024, BufferedImage.TYPE_INT_ARGB);			
			}
			Graphics g = buffer.getGraphics();

			g.clearRect(0,0,1024,1024);
			g.translate(50, 50);

			for (int x=0;x<editor.getWidth();x++) {
				for (int y=0;y<editor.getHeight();y++) {
					g.drawImage(tiles[editor.getTile(x, y)],x*32,y*32,null);
					if (editor.getUnitInNode(x, y) != null && editor.getUnitInNode(x, y).isAlive() == true) {
						g.drawImage(tiles[editor.getUnitInNode(x, y).getimageloc()],x*32,y*32,null);
					}				
				}				
			}

			graphics.drawImage(buffer, 0, 0, null);
		}	
	}
	
	/**
	 * This allows us to change the tile in the cell we click
	 */

	private void handleMousePressed(int x, int y) {

		x = x - 63;
		y = y - 84;
		x = x / 32;
		y = y / 32;

		if ((x < 0) || (y < 0) || (x >= editor.getWidth()) || (y >= editor.getHeight())) {
			return;
		}

		//cycles through tile types in map on click, switching from 0 to 1, 1 to 0, etc
		else if (editor.tiletypes[x][y] == 0) {
			editor.tiletypes[x][y] = 1;
			repaint(0);
		}

		else if (editor.tiletypes[x][y] == 1) {
			editor.tiletypes[x][y] = 2;
			repaint(0);
		}

		else if (editor.tiletypes[x][y] == 2) {
			editor.tiletypes[x][y] = 3;
			repaint(0);
		}
		

		else if (editor.tiletypes[x][y] == 3) {
			editor.tiletypes[x][y] = 0;
			repaint(0);
		}	
		repaint(0);
	}
}
