package game;

/**
 * This class is used to store potential turns for the AI
 */

public class Turn implements Comparable <Turn>{
	
	/**
	 * These variables relate to a move for the AI
	 */

	public boolean turnMoved;
	public int moveSelectionX;
	public int moveSelectionY;
	public int moveTargetX;
	public int moveTargetY;
	
	/**
	 * These variables relate to a shot for the AI
	 */
	
	public boolean turnFired;
	public int shotSelectionX;
	public int shotSelectionY;
	public int shotTargetX;
	public int shotTargetY;
	
	/**
	 * These variables are used whenever the AI makes an extra move after attacking
	 */
	
	public boolean extraMove;
	public int newTargetX;
	public int newTargetY;
	public int distanceMoved;
	
	/**
	 * Used to determine whether to execute the move or attack first when actioning the AI move
	 */
	public boolean firedFirst = false;
	
	/**
	 * Build a turn without an extra move
	 */
	
	public Turn(int a, int b, int c, int d, int e, int f, int g, int h, boolean aa, boolean ab, boolean ac){
		
		this.moveSelectionX = a;
		this.moveSelectionY = b;
		this.moveTargetX = c;
		this.moveTargetY = d;
			
		this.shotSelectionX = e;
		this.shotSelectionY = f;
		this.shotTargetX = g;
		this.shotTargetY = h;
		
		this.turnMoved = aa;
		this.turnFired = ab;
		this.firedFirst = ac;
		
		this.extraMove = false;
		this.newTargetX = -1;
		this.newTargetY = -1;
		
	}
	
	/**
	 * Build a turn with an extra move
	 */
	
	public Turn(int a, int b, int c, int d, int e, int f, int g, int h, int x, int y, boolean aa, boolean ab, boolean ac, boolean ad){
		
		this.moveSelectionX = a;
		this.moveSelectionY = b;
		this.moveTargetX = c;
		this.moveTargetY = d;
			
		this.shotSelectionX = e;
		this.shotSelectionY = f;
		this.shotTargetX = g;
		this.shotTargetY = h;
		
		this.turnMoved = aa;
		this.turnFired = ab;
		this.firedFirst = ac;
		
		this.extraMove = ad;
		this.newTargetX = x;
		this.newTargetY = y;
		
	}
	
	/**
	 * Compare two turns
	 */
	
	public int compareTo(Turn t1) {
		
		if (this.turnFired == t1.turnFired) {
			return 0; 
		}
		
		else if (this.turnMoved == true && t1.turnMoved == false) {
			return 1; 
		}
			
		else return -1;

	}
	
	/**
	 * See if two turns are equal
	 */
	
	@Override
	public boolean equals(Object other){
	    if (other == null) return false;
	    if (other == this) return true;
	    if (!(other instanceof Turn))return false;
	    Turn otherMyClass = (Turn)other;
	    if (otherMyClass.moveSelectionX == -1 ? moveSelectionX!= -1 : !otherMyClass.equals(moveSelectionX)) return false;
	    if (otherMyClass.moveSelectionY == -1 ? moveSelectionY!= -1 : !otherMyClass.equals(moveSelectionY)) return false;
	    if (otherMyClass.moveTargetX == -1 ? moveTargetX!= -1 : !otherMyClass.equals(moveTargetX)) return false;
	    if (otherMyClass.moveTargetY == -1 ? moveTargetY!= -1 : !otherMyClass.equals(moveTargetY)) return false;
	    if (otherMyClass.shotSelectionX == -1 ? shotSelectionX!= -1 : !otherMyClass.equals(shotSelectionX)) return false;
	    if (otherMyClass.shotSelectionY == -1 ? shotSelectionY!= -1 : !otherMyClass.equals(shotSelectionY)) return false;
	    if (otherMyClass.shotTargetX == -1 ? shotTargetX!= -1 : !otherMyClass.equals(shotTargetX)) return false;
	    if (otherMyClass.shotTargetY == -1 ? shotTargetY!= -1 : !otherMyClass.equals(shotTargetY)) return false;	
	    System.out.println("OK!");
	    return true;
	}
}

