package pathfinder;

/**
 * This class represents a node in the pathfinder
 * Each node has a pointer to a prior node so that when a path
 * is complete, we can follow the path back
 */

public class Node implements Comparable {

	public int x;
	public int y;

	/**
	 * The path cost for the node
	 */
	public float cost;
	public Node parent;
	
	/** 
	 * The heuristic cost of this node 
	 */
	public float heuristic;
	public int depth;
	
	public Node(int x, int y) {
		this.x = x;
		this.y = y;
	}
	
	public int setParent(Node parent) {
		depth = parent.depth + 1;
		this.parent = parent;
		
		return depth;
	}
	
	public int compareTo(Object other) {
		Node o = (Node) other;
		
		float f = heuristic + cost;
		float of = o.heuristic + o.cost;
		
		if (f < of) {
			return -1;
		} else if (f > of) {
			return 1;
		} else {
			return 0;
		}
	}
}