package players;

import java.util.*;

import units.UnitInterface;
import game.*;


public class AI_EASY_Standard implements PlayerInterface {

	/**
	 * These variables satisfy the Interface, but aren't used in
	 * this class directly
	 */

	private boolean moved = false;
	private boolean fired = false;
	private boolean movedOnce = false;

	private int selectedx = -1;
	private int selectedy = -1;
	private int lastFindx = -1;
	private int lastFindY = -1;
	private int moveDistanceTracker;

	private static int moverange = 1;
	private boolean isAI = true;
	private int playerID = 0;

	boolean winState = false;


	/**
	 * Used to instantiate the class, since AI player wont always
	 * be red or wont always be blue
	 * 
	 * @param i: red = 0, blue = 1
	 */

	public AI_EASY_Standard(int i) {
		playerID = i;
	}


	/**
	 * The variables are used to terminate the search after
	 * a given time
	 */

	long timeStart;
	long timeNow;
	long timeCheck;

	game.AIPopup popupBox = new game.AIPopup();
	game.EndGameWindow gameOver;
	boolean reSearch = true;
	boolean maxWin = false;

	/** 
	 * This class calls the function that computes the AI move
	 * It is seperate so that we can pass the returned move
	 * to another class, rather than passing the variable 
	 * that sets the returned move, and not using the returned
	 * value
	 */

	public void chooseMove(Game game) {


		Turn bestMove = chooseMoveAI(game);
		//System.out.println("here");
		swapPlayer(game, bestMove);}


	/** 
	 * This class searches for the best AI moves
	 * @param game: the game instance running
	 * @return: the best move the AI can make
	 */

	public Turn chooseMoveAI(Game game) {

		timeStart = System.currentTimeMillis();

		/**
		 * Makes a popup box appear whilst the AI is thinking
		 * It's position is relative to the center of the game window
		 */

		popupBox.setVisible(true);
		popupBox.setLocationRelativeTo(game.getLayeredPane());

		GameMap state = game.getMap();
		List<Turn> AIMoves = Game.getLegalMoves(state, false);
		System.out.println("AI MOVES: " + AIMoves.size());
		Turn bestMove = null;
		double bestScore = 0;

		for (int myDepth = 0; ; myDepth++){

			//we found a win state on the previous search so no need to continue searching
			if (maxWin == true) { System.out.println("Break due to max win ");break;}

			timeNow = System.currentTimeMillis();
			timeCheck = timeNow - timeStart;
			if (timeCheck > 5000 || myDepth >= 2) {
				System.out.println("Break due to timeout ");
				break; }
			//System.out.println("Depth is " + myDepth);	

			/**
			 * On our first search we don't have a previous search to set an estimate from, so aspAlpha/Beta are set to default 
			 * values
			 * @aspAlpha is the estimate of alpha
			 * @aspBeta is the estimate of beta
			 * @window is the size of the aspiration window
			 */

			int aspAlpha = Integer.MIN_VALUE;					
			int aspBeta = Integer.MAX_VALUE;				
			int window = 2000;


			if (bestScore != 0){

				/**
				 * Once a best score has been found at a previous depth, the bestScore at the previous depth can be used to estimate our bestScore for
				 * the next depth
				 */

				aspAlpha = (int) bestScore - window;
				aspBeta = (int) bestScore + window;
			}

			bestScore = 0;

			for (Turn t: AIMoves) {
				GameMap next = game.getNextState(state, t);
				next.turnUsed = t;
				int score;

				score = getMinScore(next, myDepth, aspAlpha, aspBeta);

				if (bestMove == null || score > bestScore) {
					bestMove = t;
					bestScore = score;
				}
			}

			if ((bestScore <= aspAlpha || bestScore >= aspBeta) && reSearch == true) {
				
				timeNow = System.currentTimeMillis();
				timeCheck = timeNow - timeStart;
				if (timeCheck > 5000) {
					System.out.println("Break due to timeout in re search");break; }
				bestScore = 0;
				bestMove = null;
				AIMoves = Game.getLegalMoves(state, true);
				aspAlpha = Integer.MIN_VALUE;
				aspAlpha = Integer.MAX_VALUE;

				for (Turn m : AIMoves) {
					GameMap next = game.getNextState(state, m);
					int score;

					// re search with default alpha/beta values	

					score = getMinScore(next, myDepth, aspAlpha, aspBeta);	

					if (bestMove == null || score > bestScore) {
						bestMove = m;
						bestScore = score;
					}	
				}														
				reSearch = true;
			}
		}
		return bestMove;	
	}

	/**
	 * Check what the worst possible counter move by the opposing player is.
	 * 
	 */

	private int getMinScore(GameMap s, int depth, double alpha, double beta) {
		if (depth == 0 || Game.isFinal(s)){
			if (Game.isFinal(s) && Game.isWinning(s)){
				reSearch = false;
			}
			int result = evaluateState(s);
			return result;
		}

		List<Turn> opponentMoves = Game.getLegalMovesMin(s, false);		
		int res = Integer.MAX_VALUE;

		for (Turn turn: opponentMoves){

			GameMap next = Game.getNextState(s, turn);
			next.turnUsed = turn;
			int score = getMaxScore(next, depth-1, alpha, beta);
			res = Math.min(res, score);
			beta = Math.min(beta, score);
			if (beta <= alpha){
				break;}
		}
		return res;
	}

	/** 
	 * Check what the best possible move the AI can make is
	 * 
	 */

	private int getMaxScore(GameMap s, int depth, double alpha, double beta) {
		if (depth == 0 || Game.isFinal(s)){
			if (Game.isFinal(s) && Game.isLosing(s)){
				reSearch = false;
				maxWin = true;}
			int result = evaluateState(s);
			return result;
		}

		List<Turn> AIMoves = Game.getLegalMoves(s, false);
		int res = Integer.MIN_VALUE;

		for (Turn turn: AIMoves){
			GameMap next = Game.getNextState(s,turn);
			next.turnUsed= turn;
			int score = getMinScore(next, depth-1, alpha, beta);
			res = Math.max(res, score);
			alpha = Math.max(alpha, score);
			if (beta <=alpha){
				break;}
		}
		return res;
	}

	public int evaluateStateMin(GameMap s) {
		int score = 5000;
		if (Game.isLosing(s))
			score+= 10000;
		if (Game.isWinning(s))
			score-=10000;   


		if (playerID == 0) {
			if (s.blueUnitCount > s.redUnitCount){
				score += 5000;
			}
			if (s.redUnitCount > s.blueUnitCount){
				score -= 5000;
			}		
		}

		else if (playerID == 1) {
			if (s.redUnitCount > s.blueUnitCount){
				score += 5000;
			}
			if (s.blueUnitCount > s.redUnitCount){
				score -= 5000;
			}
		}


		for (int i = 0; i < s.getWidth();i++) {
			for (int j = 0; j < s.getHeight();j++) {
				//if there's a unit here and not on our team
				if (s.getUnitInNode(i,j) != null){
					if (s.getUnitInNode(i, j).getTeamID() != playerID){
						UnitInterface myunit = s.getUnitInNode(i,j);

						if (myunit.getimageloc() == 10 || myunit.getimageloc() == 11){
							score += 1000;
						}

						if (myunit.getimageloc() == 8 || myunit.getimageloc() == 9){
							score += 750;
						}
						if (myunit.getimageloc() == 6 || myunit.getimageloc() == 7){
							score += 500;
						}
						if (myunit.getimageloc() == 4 || myunit.getimageloc() == 5){
							score += 250;
						}
					}
				}
			}
		}

		int random = (-50) + (int)(Math.random() * ((50 - (-50)) + 1));
		score += random;
		//System.out.println("Random is: " + random);

		//System.out.println("Min Score is: " + score);
		return score;
	}
	
	/**
	 * This player evaluates most units as similarly useful but is balanced between
	 * losing points for playing badly and gaining them for being at an advantage
	 */

	public int evaluateState(GameMap s) {
		int score = 5000;
		if (Game.isLosing(s))
			score+= 10000;
		if (Game.isWinning(s))
			score-=10000;   


		if (playerID == 0) {
			if (s.blueUnitCount < s.redUnitCount){
				score += 5000;
			}
			if (s.redUnitCount < s.blueUnitCount){
				score -= 5000;
			}

		}

		else if (playerID == 1) {
			if (s.redUnitCount < s.blueUnitCount){
				score += 5000;
			}
			if (s.blueUnitCount < s.redUnitCount){
				score -= 5000;
			}
		}

		for (int i = 0; i < s.getWidth();i++) {
			for (int j = 0; j < s.getHeight();j++) {
				//if there's a unit here and on our team
				if (s.getUnitInNode(i,j) != null){
					if (s.getUnitInNode(i, j).getTeamID() == playerID){
						UnitInterface myunit = s.getUnitInNode(i,j);

						if (myunit.getimageloc() == 10 || myunit.getimageloc() == 11){
							score += 500;
						}

						if (myunit.getimageloc() == 8 || myunit.getimageloc() == 9){
							score += 500;
						}
						if (myunit.getimageloc() == 6 || myunit.getimageloc() == 7){
							score += 500;
						}
						if (myunit.getimageloc() == 4 || myunit.getimageloc() == 5){
							score += 250;
						}
					}
				}
			}
			//System.out.println("AI Score: " + score);
		}
		//System.out.println("Max Score is: " + score);
		int random = (-1000) + (int)(Math.random() * ((1000 - (-1000)) + 1));
		//System.out.println("Random is: " + random);
		score += random;
		return score;

	}

	public void swapPlayer(Game game, Turn bestMove) {

		System.out.println("\n Move sel x: " + bestMove.moveSelectionX);
		System.out.println("\n Move sel y: " + bestMove.moveSelectionY);
		System.out.println("\n Move tgt x: " + bestMove.moveTargetX);
		System.out.println("\n Move tgt y: " + bestMove.moveTargetY);

		System.out.println("\n Shot sel x: " + bestMove.shotSelectionX);
		System.out.println("\n Shot sel y: " + bestMove.shotSelectionY);
		System.out.println("\n Shot tgt x: " + bestMove.shotTargetX);
		System.out.println("\n Shot tgt y: " + bestMove.shotTargetY);

		System.out.println("\n Extra move x: " + bestMove.newTargetX);
		System.out.println("\n Extra move x: " + bestMove.newTargetY);


		popupBox.dispose();
		game.doAIMove(bestMove, game.getMap());

		/**
		 * each units position needs to be updated so that the can properly clone the game state
		 */

		game.getMap().builtUnits = new ArrayList<UnitInterface>();

		for (int i = 0; i < game.getMap().getWidth(); i++){
			for (int j = 0; j < game.getMap().getHeight(); j++){
				if (game.getMap().getUnitInNode(i,j) != null && game.getMap().getUnitInNode(i,j).isAlive() == true){
					UnitInterface unit = game.getMap().getUnitInNode(i,j);
					//System.out.println("Unit found in " + i + ", " + j);
					unit.setCurrentPosX(i);
					unit.setCurrentPosY(j);
					game.getMap().builtUnits.add(unit);
					//System.out.println("Units currentpos is now " + game.getMap().getUnitInNode(i,j).getCurrentPosX() + ", " + game.getMap().getUnitInNode(i,j).getCurrentPosY());
				}
			}
		}


		for (UnitInterface u: game.getMap().builtUnits){
			System.out.println("Unit u has current position " + u.getCurrentPosX() + ", " + u.getCurrentPosY());
		}

		System.out.println("red units: " + game.getMap().redUnitCount);
		System.out.println("blue units: " + game.getMap().blueUnitCount);



		System.gc();
		System.out.println("DONE!");


		game.getGlassPane().setVisible(true);

		setSelectedx(-1);
		setSelectedy(-1);
		setLastFindx(-1);
		setFired(false);
		setMoved(false);	

		if (game.getTurnID() == 0) { game.setTurnID(1);}
		else if (game.getTurnID() == 1) { game.setTurnID(0);}

		game.turnCount += 1;
		System.out.println("Turns taken =" + game.turnCount);
		System.out.println("Turn Limit is =" + game.maxTurns);
		if (game.maxTurns != -1 && game.turnCount >= game.maxTurns){
			gameOver = new game.EndGameWindow(game.turnCount, game.getMap().redUnitCount, game.getMap().blueUnitCount, game.initialRed, game.initialBlue);
			gameOver.setVisible(true);
			game.dispose();
		}

		if (game.getMap().redUnitCount == 0 || game.getMap().blueUnitCount == 0) {
			gameOver = new game.EndGameWindow(game.turnCount, game.getMap().redUnitCount, game.getMap().blueUnitCount, game.initialRed, game.initialBlue);
			gameOver.setVisible(true);
			//playOn = false;
			game.dispose();
		}

		//these only execute if game's not over

		else if (game.getCurrentPlayer() == game.getPlayer1()) {
			game.currentPlayer = game.getPlayer2();
			game.currentPlayer.chooseMove(game);
		}
		else if (game.getCurrentPlayer() == game.getPlayer2()) {
			game.currentPlayer = game.getPlayer1();
			game.currentPlayer.chooseMove(game);


		}	
		game.repaint(0);
	}


	@Override
	public int getPlayerID() {
		return playerID;
	}

	@Override
	public void setPlayerID(int x) {
		playerID = x;		
	}

	@Override
	public boolean isAI() {
		return isAI;
	}

	//copied from mpgame

	public int getSelectedx() {
		return selectedx;
	}

	public void setSelectedx(int selectedx) {
		this.selectedx = selectedx;
	}

	public int getSelectedy() {
		return selectedy;
	}

	public void setSelectedy(int selectedy) {
		this.selectedy = selectedy;
	}

	public int getLastFindx() {
		return lastFindx;
	}

	public void setLastFindx(int lastFindx) {
		this.lastFindx = lastFindx;
	}

	public int getLastFindY() {
		return lastFindY;
	}

	public void setLastFindY(int lastFindY) {
		this.lastFindY = lastFindY;
	}

	public boolean isFired() {
		return fired;
	}

	public void setFired(boolean fired) {
		this.fired = fired;
	}

	public boolean isMoved() {
		return moved;
	}

	public void setMoved(boolean moved) {
		this.moved = moved;
	}

	public boolean isMovedOnce() {
		return movedOnce;
	}

	public void setMovedOnce(boolean movedOnce) {
		this.movedOnce = movedOnce;
	}

	public int getMoveDistanceTracker() {
		return moveDistanceTracker;
	}

	public void setMoveDistanceTracker(int moveDistanceTracker) {
		this.moveDistanceTracker = moveDistanceTracker;
	}

	public int getMoverange() {
		return moverange;
	}
	public void setMoverange(int moverange) {
		this.moverange = moverange;
	}

	@Override
	public void swapPlayer(Game game) {
		// TODO Auto-generated method stub

	}
}
